# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformation modules that convert or pre/post process data structures."""

import dataclasses
import functools
import re
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Tuple
from dinosaur import coordinate_systems
from dinosaur import pytree_utils
from dinosaur import sigma_coordinates
from dinosaur import typing

import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import filters
import numpy as np
import tensorflow_probability.substrates.jax as tfp


KeyWithCosLatFactor = typing.KeyWithCosLatFactor
TransformModule = typing.TransformModule


@gin.register
class EmptyTransform(hk.Module):
  """Transform returns an empty dict."""

  def __init__(self, *args, name: Optional[str] = None):
    del args  # unused.
    super().__init__(name=name)

  def __call__(self, inputs: ...) -> typing.Pytree:
    return {}


@gin.register
class IdentityTransform(hk.Module):
  """Transform does not modify inputs."""

  def __init__(self, *args, name: Optional[str] = None, **kwargs):
    del args, kwargs  # unused.
    super().__init__(name=name)

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return inputs


@gin.register
class ShiftAndNormalize(hk.Module):
  """Transforms inputs by shifting and normalizing values by `shifts/scales`."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      shifts: typing.Pytree,
      scales: typing.Pytree,
      features_to_exclude: Sequence[str] = tuple(),
      global_scale: Optional[float] = None,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.shifts = shifts
    if global_scale is not None:
      scales = jax.tree_util.tree_map(lambda x: x * global_scale, scales)
    self.scales = scales

  def __call__(self, inputs):
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    shifts = pytree_utils.replace_with_matching_or_default(
        inputs, self.shifts, default=None, check_used_all_replace_keys=False)
    scales = pytree_utils.replace_with_matching_or_default(
        inputs, self.scales, default=None, check_used_all_replace_keys=False)
    # if shifts/scales have missing values present in `inputs`, we insert `None`
    # for the default. If corresponding `inputs` is not `None`, this will raise
    # an error, as expected. This works because tree_map skips `None` values in
    # the first argument, as long as all dictionary keys match.
    result = jax.tree_util.tree_map(
        lambda x, y, z: (x - y) / z, inputs, shifts, scales)
    return from_dict_fn(result)


@gin.register
class InverseShiftAndNormalize(hk.Module):
  """Inverse of the `ShiftAndNormalize` for the same `shifts/scales`."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      shifts: typing.Pytree,
      scales: typing.Pytree,
      global_scale: Optional[float] = None,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.shifts = shifts
    if global_scale is not None:
      scales = jax.tree_util.tree_map(lambda x: x * global_scale, scales)
    self.scales = scales

  def __call__(self, inputs):
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    shifts = pytree_utils.replace_with_matching_or_default(
        inputs, self.shifts, default=None, check_used_all_replace_keys=False)
    scales = pytree_utils.replace_with_matching_or_default(
        inputs, self.scales, default=None, check_used_all_replace_keys=False)
    # if shifts/scales have missing values present in `inputs`, we insert `None`
    # for the default. If corresponding `inputs` is not `None`, this will raise
    # an error, as expected. This works because tree_map skips `None` values in
    # the first argument, as long as all dictionary keys match.
    result = jax.tree_util.tree_map(
        lambda x, y, z: (None if x is None else x * z + y),
        inputs,
        shifts,
        scales,
        is_leaf=lambda x: x is None,
    )
    return from_dict_fn(result)


@gin.register
class ToModalWithDivCurlTransform(hk.Module):
  """Module that converts inputs to modal replacing velocity with div/curl."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    if 'u' not in inputs or 'v' not in inputs:
      raise ValueError('Inputs to ToModalWithDivCurlTransform must include `u, '
                       f'v`, got keys: {inputs.keys()}')
    sec_lat = 1 / self.coords.horizontal.cos_lat
    u, v = inputs.pop('u'), inputs.pop('v')
    # here u,v stand for velocity / cos(lat), but the cos(lat) is cancelled in
    # divergence and curl operators below.
    inputs['u'] = u * sec_lat
    inputs['v'] = v * sec_lat
    to_modal_fn = lambda x: (self.coords.horizontal.to_modal(x)  # pylint: disable=g-long-lambda
                             if x is not None else None)
    modal_outputs = jax.tree_util.tree_map(to_modal_fn, inputs)
    u, v = modal_outputs.pop('u'), modal_outputs.pop('v')
    modal_outputs['divergence'] = self.coords.horizontal.div_cos_lat((u, v))
    modal_outputs['vorticity'] = self.coords.horizontal.curl_cos_lat((u, v))
    return modal_outputs


@gin.register
class ToModalDiffOperators(hk.Module):
  """Module that returns grad and laplacian features of inputs fields.

  To avoid accidental accumulation of the cos(lat) factors, features must be
  keyed using typing.KeyWithCosLatFactor namedtuple.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords

  def __call__(
      self,
      inputs: Mapping[typing.KeyWithCosLatFactor, typing.Array],
  ) -> Mapping[typing.KeyWithCosLatFactor, typing.Array]:
    features = {}
    for k, value in inputs.items():
      name, cos_lat_order = k.name, k.factor_order
      d_value_dlon, d_value_dlat = self.coords.horizontal.cos_lat_grad(value)
      laplacian_value = self.coords.horizontal.laplacian(value)
      dlon_key = typing.KeyWithCosLatFactor(name + '_dlon', cos_lat_order + 1)
      dlat_key = typing.KeyWithCosLatFactor(name + '_dlat', cos_lat_order + 1)
      del2_key = typing.KeyWithCosLatFactor(name + '_del2', cos_lat_order)
      features[dlon_key] = d_value_dlon
      features[dlat_key] = d_value_dlat
      features[del2_key] = laplacian_value
    return features


@gin.register
class ModalToNodalTransform(hk.Module):
  """Transform that converts modal inputs to nodal representation."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return self.coords.horizontal.to_nodal(inputs)


@gin.register
class NodalToModalTransform(hk.Module):
  """Transform that converts nodal inputs to modal representation."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return self.coords.horizontal.to_modal(inputs)


@gin.register
class ClipTransform(hk.Module):
  """Transform that clips highest total wavenumber in inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      wavenumbers_to_clip: int = 1,
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.coords = coords
    self.wavenumbers_to_clip = wavenumbers_to_clip

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    return self.coords.horizontal.clip_wavenumbers(
        inputs, self.wavenumbers_to_clip
    )


@gin.register
class SinhArcsinhTransform(hk.Module):
  """Transform that magnifies or suppresses inputs with large magnitudes."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      tailweight: float = 1.0,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.tailweight = tailweight

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    f = tfp.bijectors.SinhArcsinh(tailweight=self.tailweight).forward
    return jax.tree_util.tree_map(f, inputs)


@gin.register
class NondimensionalizeTransform(hk.Module):
  """Transform that nondimensionalizes inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      input_coords: coordinate_systems.CoordinateSystem,
      inputs_to_units_mapping: Dict[str, str],
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del coords, dt, aux_features, input_coords  # unused.
    super().__init__(name=name)
    self.inputs_to_units_mapping = inputs_to_units_mapping
    self.nondimensionalize = physics_specs.nondimensionalize

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    inputs_to_units_mapping = pytree_utils.replace_with_matching_or_default(
        inputs, self.inputs_to_units_mapping, default=None,
        check_used_all_replace_keys=False,
    )
    nondim_fn = lambda x, y: self.nondimensionalize(x * typing.Quantity(y))
    result = jax.tree_util.tree_map(nondim_fn, inputs, inputs_to_units_mapping)
    return from_dict_fn(result)


@gin.register
class RedimensionalizeTransform(hk.Module):
  """Transform that redimensionalizes inputs."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      output_coords: coordinate_systems.CoordinateSystem,
      inputs_to_units_mapping: Dict[str, str],
      name: Optional[str] = None,
  ):
    """See `time_integration.exponential_filter` for details."""
    del coords, dt, aux_features, output_coords  # unused.
    super().__init__(name=name)
    self.inputs_to_units_mapping = inputs_to_units_mapping
    self.dimensionalize = physics_specs.dimensionalize

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    inputs_to_units_mapping = pytree_utils.replace_with_matching_or_default(
        inputs, self.inputs_to_units_mapping, default=None,
        check_used_all_replace_keys=False,
    )
    dim_fn = lambda x, y: self.dimensionalize(x, typing.Quantity(y)).m
    result = jax.tree_util.tree_map(dim_fn, inputs, inputs_to_units_mapping)
    return from_dict_fn(result)


@gin.register
class SequentialTransform(hk.Module):
  """Transform module that combines multiple transforms applied sequentially."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      transform_modules: Sequence[TransformModule],
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.transform_fns = [module(coords, dt, physics_specs, aux_features)
                          for module in transform_modules]

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    for transform_fn in self.transform_fns:
      inputs = transform_fn(inputs)
    return inputs


@gin.register
class LevelScale(hk.Module):
  """Transforms inputs by scaling different vertical levels."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      scales: Sequence[float],
      keys_to_scale: Sequence[str] = tuple(),
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.scale_fn = functools.partial(
        coordinate_systems.scale_levels_for_matching_keys,
        scales=np.asarray(scales),
        keys_to_scale=keys_to_scale)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.scale_fn(inputs)


@gin.register
class InverseLevelScale(hk.Module):
  """Transforms inputs by inverse scaling different vertical levels."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      scales: Sequence[float],
      keys_to_scale: Sequence[str] = tuple(),
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.scale_fn = functools.partial(
        coordinate_systems.scale_levels_for_matching_keys,
        scales=1/np.asarray(scales),
        keys_to_scale=keys_to_scale)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.scale_fn(inputs)


@gin.register
class HardClip(hk.Module):
  """Transforms inputs by hard clipping inputs to (-max_value, max_value)."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      max_value: float,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    super().__init__(name=name)
    self.clip_fn = functools.partial(
        jnp.clip, min=-max_value, max=max_value)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return jax.tree_util.tree_map(self.clip_fn, inputs)


@gin.register
class SoftClip(hk.Module):
  """Transforms inputs by clipping values to a range with smooth boundaries.

  Attributes:
    coords: horizontal and vertical descritization.
    dt: time step of the model.
    physics_specs: object describing the scales and physical constants.
    aux_features: dictionary holding static features that the model may use.
    max_value: specifies the range (-max_value, max_value) of return values.
    hinge_softness: controls the softness of the smoothing at the boundaries;
      values outside of the max_value range are mapped into intervals of width
      approximately `log(2) * hinge_softness` on the interior of each boundary.
    name: optional name of the module.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      max_value: float,
      hinge_softness: float = 1.0,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused.
    if max_value < 0 or hinge_softness < 0:
      raise ValueError('max_value and hinge_softness must be positive, '
                       f'{max_value=}, {hinge_softness=}')
    super().__init__(name=name)
    low = -max_value
    high = max_value
    hinge = hinge_softness
    softplus_fn = lambda x: hinge * jax.nn.softplus(x / hinge)
    self.clip_fn = lambda x: (  # pylint: disable=g-long-lambda
        -softplus_fn(high - low - softplus_fn(x - low)) *
        (high - low) / (softplus_fn(high - low)) + high)

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return jax.tree_util.tree_map(self.clip_fn, inputs)


@gin.register
class ToModalDiffOperatorsWithFiltering(hk.Module):
  """Module that returns filtered grad and laplacian features of inputs fields.

  To avoid accidental accumulation of the cos(lat) factors, features must be
  keyed using typing.KeyWithCosLatFactor namedtuple.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      filter_attenuations: Tuple[float, ...] = tuple(),
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.attenuations = filter_attenuations
    feature_filters = []
    for attenuation in filter_attenuations:
      feature_filters.append(
          filters.DataExponentialFilter(
              coords, dt, physics_specs, aux_features,
              order=1, attenuation=attenuation))
    self.feature_filters = feature_filters

  def __call__(
      self,
      inputs: Mapping[KeyWithCosLatFactor, typing.Array],
  ) -> Mapping[KeyWithCosLatFactor, typing.Array]:
    features = {}
    for k, value in inputs.items():
      name, cos_lat_order = k.name, k.factor_order
      for filter_fn, att in zip(self.feature_filters, self.attenuations):
        filtered_value = filter_fn(value)
        d_value_dlon, d_value_dlat = self.coords.horizontal.cos_lat_grad(
            filtered_value)
        laplacian_value = self.coords.horizontal.laplacian(filtered_value)
        # since gradient values picked up cos_lat factor we increment the
        # corresponding key. This factor is adjusted at the caller level.
        dlon_key = KeyWithCosLatFactor(
            name + f'_dlon_{att}', cos_lat_order + 1, att)
        dlat_key = KeyWithCosLatFactor(
            name + f'_dlat_{att}', cos_lat_order + 1, att)
        del2_key = KeyWithCosLatFactor(
            name + f'_del2_{att}', cos_lat_order, att)
        features[dlon_key] = d_value_dlon
        features[dlat_key] = d_value_dlat
        features[del2_key] = laplacian_value
    return features


@gin.register
class TruncateSigmaLevels(hk.Module):
  """Transform module that truncates vertical levels for specified variables."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      sigma_ranges: dict[str, Tuple[float, float]],
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    del dt, physics_specs, aux_features  # unused.
    self.sigma_ranges = sigma_ranges
    self.sigma_levels = coords.vertical.centers

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    """Returns `inputs` where only specified levels are retained."""

    def _slice_fn(x, sigma_range):
      """Returns `x` sliced to include values in `sigma_range`."""
      sigma_min_slice, sigma_max_slice = sigma_range
      lower_index = np.argmax((self.sigma_levels - sigma_min_slice) > 0)
      if sigma_max_slice > np.max(self.sigma_levels):
        upper_index = len(self.sigma_levels)
      else:
        upper_index = np.argmin((self.sigma_levels - sigma_max_slice) < 0)
      return x[slice(lower_index, upper_index), ...]

    def recurse_and_replace(x: dict[str, Any],
                            y: dict[str, Any],
                            default=None) -> dict[str, Any]:
      """Copy x, setting leaf values to `default` or value from y if keys match."""
      return {
          k: (
              y.get(k, default)
              if not isinstance(v, dict)
              else recurse_and_replace(v, y, default)
          )
          for k, v in x.items()
      }

    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    sigma_ranges_extended = recurse_and_replace(
        inputs, self.sigma_ranges, default=(0, 1)
    )
    outputs = jax.tree_util.tree_map(_slice_fn, inputs, sigma_ranges_extended)
    return from_dict_fn(outputs)


@gin.register
class TakeSurfaceAdjacentSigmaLevel(hk.Module):
  """Transform module that retains only the vertical level nearest to Earth surface for all variables."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    del coords, dt, physics_specs, aux_features  # unused.

  def __call__(self, inputs: typing.PyTreeState) -> typing.PyTreeState:
    """Returns `inputs` where only last sigma level is retained."""

    def _slice_fn(x):
      return x[slice(-1, None), ...]

    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    outputs = jax.tree_util.tree_map(_slice_fn, inputs)
    return from_dict_fn(outputs)


@gin.register
@dataclasses.dataclass
class FeatureSelector:
  """Features transform that retains items whose keys match against regex.

  Attributes:
    regex_patterns: regular expression pattern that specifies the set of keys
      from `inputs` that will be returned by __call__ method.
  """
  regex_patterns: str

  def __call__(
      self,
      inputs: Dict[str, typing.Array],
  ) -> Dict[str, typing.Array]:
    outputs = {}
    for k, v in inputs.items():
      if re.fullmatch(self.regex_patterns, k):
        outputs[k] = v
    return outputs


@gin.register
class BroadcastTransform:
  """Features transform that broadcasts all features."""

  def __init__(self, *args, **kwargs):
    del args, kwargs  # unused.

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    leaves, tree_def = jax.tree_util.tree_flatten(inputs)
    leaves = jnp.broadcast_arrays(*leaves)
    return jax.tree_util.tree_unflatten(tree_def, leaves)


@gin.register
class SquashLevelsTransform:
  """Transform that "squashes" values of inputs depending on their sigma level.

  Multiplies inputs by the piecewise linear values used to "squash" inputs
  by sigma level. See function χ definition at: http://screen/5V3jzU7ZFA4vVJP

  The squash is paramtereizaed by low_cutoffs and high_cutoffs.
  On Palmer 2009 (http://shortn/_56HCcQwmSS) page 4, the cutoffs for
  perturbations are given. Below are translated to sigma levels values:
    low_cutoffs: (100hPa, 50hPa)
    low_cutoffs: (0.05, 0.1),
    high_cutoffs: (1300m, 300m)
    high_cutoffs: (0.86, 0.965)

  Inputs that have a singleton or no level dimension are assumed defined at
  the highest value of sigma ("surface level").

  Attributes:
    coords: horizontal and vertical descritization.
    dt: time step of the model.
    physics_specs: object describing the scales and physical constants.
    aux_features: dictionary holding static features that the model may use.
    low_cutoffs: σ=low_cutoffs[0] is when χ starts linearly increasing from 0.
      σ=low_cutoffs[1] is when χ levels out at 1
    high_cutoffs: σ=high_cutoffs[0] is when χ starts linearly decreasing from 1.
      σ=high_cutoffs[1] is when χ reaches 0.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      low_cutoffs: Sequence[float] = (0.05, 0.1),
      high_cutoffs: Sequence[float] = (0.86, 0.965),
  ):
    del dt, physics_specs, aux_features  # unused.
    if not isinstance(coords.vertical, sigma_coordinates.SigmaCoordinates):
      raise ValueError(f'Cannot apply sigma_squash on {coords.vertical=}')
    sigma = coords.vertical.centers
    if len(low_cutoffs) != 2:
      raise ValueError(f'{len(low_cutoffs)=} but should have been 2.')
    if len(high_cutoffs) != 2:
      raise ValueError(f'{len(high_cutoffs)=} but should have been 2.')

    low_func = (sigma - low_cutoffs[0]) / (low_cutoffs[1] - low_cutoffs[0])
    high_func = (high_cutoffs[1] - sigma) / (high_cutoffs[1] - high_cutoffs[0])

    # lower_bound is a function equal to the squasher between
    # low_cutoffs[0] and high_cutoffs[1].
    # It becomes negative outside that range.
    lower_bound = np.minimum(1., np.minimum(low_func, high_func))
    self._sigma_squash = np.maximum(0., lower_bound)[:, np.newaxis, np.newaxis]

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    def squash_per_level_only(x):
      shape = jnp.shape(x)
      ndim = len(shape)
      if ndim >= 3 and shape[-3] > 1:  # If defined per-level
        return x * self._sigma_squash
      elif ndim in {2, 3}:
        return x * self._sigma_squash[-1]   # If defined at surface level
      else:
        return x
    return jax.tree_util.tree_map(squash_per_level_only, inputs)


@gin.register
def add_prefix(features: dict[str, Any], prefix: str) -> dict[str, Any]:
  """Adds prefix to keys in features."""
  return {prefix + k: v for k, v in features.items()}


def straight_through(
    f: Callable[[typing.Array], typing.Array],
) -> Callable[[typing.Array], typing.Array]:
  """Straight-through estimator of `func`.

  The "straight-through" estimator is a trick that fools auto-diff into
  assigning a constant gradient (≡ 1) to a function.
  See http://shortn/_kRQjMbF2QF

  Args:
    f: Callable mapping arrays to arrays. May be non-differentiable.

  Returns:
    g: Function g such that g(x) ≡ f(x) and g'(x) ≡ 1.
  """
  def straight_through_f(x):
    zero = x - jax.lax.stop_gradient(x)
    return zero + jax.lax.stop_gradient(f(x))
  return straight_through_f



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that transform data between pytrees."""

from typing import Callable, Optional, Sequence
from dinosaur import pytree_utils
from dinosaur import typing
import gin
import haiku as hk
import jax

from neuralgcm import transforms


Array = typing.Array
Tower = Callable[[int], Callable[..., Array]]
MappingModule = Callable[[typing.Pytree], typing.Pytree]


@gin.register(denylist=['output_shapes'])
class NodalMapping(hk.Module):
  """Maps the pytree of nodal features to a pytree of specified structure.

  This module packs the pytree into a single array of shape (n, lon, lat),
  passes it to a NN tower, and unpacks the result into a pytree with the
  structure of output_shapes, typically (m, lon, lat).
  """

  def __init__(
      self,
      output_shapes: typing.Pytree,
      tower_factory: Tower = gin.REQUIRED,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    feature_axis = -3  # default column axis.
    output_size = sum([x[feature_axis]
                       for x in jax.tree_util.tree_leaves(output_shapes)])
    # tower preserves the last two spatial dimensions.
    self.tower = tower_factory(output_size)
    self.output_shapes = output_shapes
    self.feature_axis = feature_axis

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    array = pytree_utils.pack_pytree(inputs, self.feature_axis)
    if array.ndim != 3:
      raise ValueError(f'Expected input array with ndim=3, got {array.shape=}')
    outputs = self.tower(array)
    if outputs.ndim != 3:
      raise ValueError(f'Expected outputs with ndim=3, got {outputs.shape=}')
    return pytree_utils.unpack_to_pytree(
        outputs, self.output_shapes, self.feature_axis)


@gin.register(denylist=['output_shapes'])
class NodalVolumeMapping(hk.Module):
  """Maps the pytree of nodal volume features to a pytree of given structure.

  This module stacks the input pytree into an array of shape
  (channel, level, lon, lat), passes it to a NN tower.  The output from the NN
  is expected to have shape (n, level, lon, lat), and gets unpacked to a pytree
  with the structure of output_shapes, e.g.
    output_shapes = {
        'var_1': jnp.asarray((level, lon, lat)),
        'var_2': jnp.asarray((level, lon, lat)),
        ...,
        'var_n': jnp.asarray((level, lon, lat)),
    }

  The leaves of the input pytree must have the same shape, e.g. (1, lon, lat) or
  (level, lon, lat). To mix shapes, broadcast before passing to the mapping.
  """

  def __init__(
      self,
      output_shapes: typing.Pytree,
      tower_factory: Tower = gin.REQUIRED,
      name: Optional[str] = None
  ):
    super().__init__(name=name)
    feature_axis = 0
    output_size = len(jax.tree_util.tree_leaves(output_shapes))

    # tower preserves the last two spatial dimensions.
    self.tower = tower_factory(output_size)
    self.output_shapes = output_shapes
    self.feature_axis = feature_axis

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    array = pytree_utils.stack_pytree(inputs, axis=self.feature_axis)
    if array.ndim != 4:
      raise ValueError(f'Expected input array with ndim=4, got {array.shape=}')
    outputs = self.tower(array)
    if outputs.ndim != 4:
      raise ValueError(f'Expected outputs with ndim=4, got {outputs.shape=}')
    return pytree_utils.unstack_to_pytree(
        outputs, self.output_shapes, axis=self.feature_axis
    )


@gin.register
class NodalVolumeTransformerMapping(hk.Module):
  """Maps the pytree of nodal volume features to a pytree of given structure.

  Similar to NodalVolumeMapping, but uses certain features as positional
  encoding arguments to the underlying transformer networks. Inputs are
  expected to be of shape (channel, level, lon, lat), which are split into
  encoder inputs, decoder inputs and positional encodings, which are then passed
  to transformer towers. The output of the NN is expected to have shape
  (n, level*, lon, lat), and gets unpacked to a pytree with the structure of
  output_shapes, e.g.
    output_shapes = {
        'out_1': jnp.asarray((level*, lon, lat)),
        'out_2': jnp.asarray((level*, lon, lat)),
        ...,
        'out_n': jnp.asarray((level*, lon, lat)),
    }
  Note: the output number of levels `level*` is equal to those defined by the
  `decoder_inputs_selection_module`. In case it is empty, level* == level.

  The leaves of the encoder/decoder pytrees must have the same shape, e.g.
  (1, lon, lat) or (level, lon, lat) or (level*, lon, lat). To mix shapes,
  broadcast before passing to the mapping.
  """

  def __init__(
      self,
      output_shapes: typing.Pytree,
      encoder_transformer_tower_factory: Tower = gin.REQUIRED,
      decoder_transformer_tower_factory: Tower = gin.REQUIRED,
      latent_size: int = gin.REQUIRED,
      encoder_inputs_selection_module=gin.REQUIRED,
      decoder_inputs_selection_module=transforms.EmptyTransform,
      encoder_pos_encoding_module=transforms.EmptyTransform,
      decoder_pos_encoding_module=transforms.EmptyTransform,
      name: Optional[str] = None
  ):
    super().__init__(name=name)
    feature_axis = 0
    output_size = len(jax.tree_util.tree_leaves(output_shapes))
    self.encoder_tower = encoder_transformer_tower_factory(latent_size)
    self.decoder_tower = decoder_transformer_tower_factory(output_size)
    self.output_shapes = output_shapes
    self.feature_axis = feature_axis
    self.get_encoder_inputs_fn = encoder_inputs_selection_module()
    self.get_decode_inputs_fn = decoder_inputs_selection_module()
    self.encoder_positional_encodings_fn = encoder_pos_encoding_module()
    self.decoder_positional_encodings_fn = decoder_pos_encoding_module()

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    enc_inputs = self.get_encoder_inputs_fn(inputs)
    dec_inputs = self.get_decode_inputs_fn(inputs)
    enc_array = pytree_utils.stack_pytree(enc_inputs, axis=self.feature_axis)
    dec_array = pytree_utils.stack_pytree(dec_inputs, axis=self.feature_axis)
    enc_pos_encoding = pytree_utils.stack_pytree(
        self.encoder_positional_encodings_fn(inputs), axis=self.feature_axis)
    dec_pos_encoding = pytree_utils.stack_pytree(
        self.decoder_positional_encodings_fn(inputs), axis=self.feature_axis)
    input_ndims = set(
        x.ndim
        for x in [enc_array, dec_array, enc_pos_encoding, dec_pos_encoding]
        if x is not None)
    if input_ndims != {4}:
      raise ValueError(f'Expected all inputs have ndim=4, got {input_ndims=}')
    latents = self.encoder_tower(enc_array, None, enc_pos_encoding)
    # if dec_array is None, use latents as `inputs` and provide no `latents`.
    decoder_latents = None if dec_array is None else latents
    # if dec_array is None, use `latents`, otherwise use dec_array as `inputs`.
    dec_array = dec_array if dec_array is not None else latents
    outputs = self.decoder_tower(dec_array, decoder_latents, dec_pos_encoding)
    if outputs.ndim != 4:
      raise ValueError(f'Expected outputs with ndim=4, got {outputs.shape=}')
    return pytree_utils.unstack_to_pytree(
        outputs, self.output_shapes, axis=self.feature_axis
    )


@gin.register(denylist=['output_shapes'])
class ParallelMapping(hk.Module):
  """Maps a pytree to a pytree by additively compbining multiple mappings.

  Outputs of `mappings` must be compatible with each other.
  """

  def __init__(
      self,
      output_shapes: typing.Pytree,
      mappings: Sequence[MappingModule] = gin.REQUIRED,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.mapping_fns = [m(output_shapes) for m in mappings]

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    results = [mapping_fn(inputs) for mapping_fn in self.mapping_fns]
    return jax.tree_util.tree_map(lambda *args: sum(args), *results)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Basic neural network towers for whirl/gcm codebase.

A tower is a neural network that operates identically over the last two
dimensions, i.e. (longitude, latitude).
"""
from collections import abc
from typing import Callable, Optional, Tuple
from dinosaur import typing
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import layers

Array = typing.Array
TowerFactory = typing.TowerFactory
LayerFactory = typing.LayerFactory
MLP = gin.external_configurable(hk.nets.MLP)


@gin.register(denylist=['output_size'])
class ColumnTower(hk.Module):
  """Column tower module parameterized by column_net_factory."""

  def __init__(
      self,
      output_size: int,
      column_net_factory: LayerFactory = gin.REQUIRED,
      checkpoint_tower: bool = False,
      name: Optional[str] = None,
  ):
    """Tower that maps a column_net over two spatial dimensions."""
    super().__init__(name=name)
    column_net = column_net_factory(output_size)
    vmap_last = lambda fn: hk.vmap(fn, in_axes=-1, out_axes=-1, split_rng=False)
    column_tower = vmap_last(vmap_last(column_net))
    if checkpoint_tower:
      column_tower = hk.remat(column_tower)
    self.column_tower = column_tower

  def __call__(self, inputs: Array) -> Array:
    """Applies Column tower to inputs."""
    return self.column_tower(inputs)


@gin.register(denylist=['output_size'])
class ColumnTransformerTower(ColumnTower):
  """Same as ColumnTower, but passes additional transformer inputs."""

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def __call__(
      self,
      inputs: Array,
      latents: Optional[Array] = None,
      positional_encoding: Optional[Array] = None,
  ) -> Array:
    """Applies Column tower to inputs."""
    return self.column_tower(inputs, latents, positional_encoding)


@gin.register(denylist=['output_size'])
class VerticalConvTower(hk.Module):
  """Tower that stacks up layers of Conv1D.

  input shape: [in_channel, level, lon, lat],
  output shape: [output_size, level, lon, lat].
  """

  def __init__(
      self,
      output_size: int,  # The number of channels in the last layer
      channels: abc.Sequence[int] = gin.REQUIRED,
      kernel_shape: int = gin.REQUIRED,
      with_bias: bool = True,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.relu,
      activate_final: bool = False,
      checkpoint_tower: bool = False,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.activation = activation
    self.output_size = output_size
    self.activate_final = activate_final
    self.checkpoint_tower = checkpoint_tower

    self.layers = []
    channels = list(channels) + [self.output_size]
    for channels_i in channels:
      self.layers.append(layers.ConvLevel(
          output_channels=channels_i,
          kernel_shape=kernel_shape,
          with_bias=with_bias))

  def net(self, inputs: Array) -> Array:
    out = inputs
    num_layers = len(self.layers)
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < (num_layers - 1) or self.activate_final:
        out = self.activation(out)
    return out

  def __call__(self, inputs: Array) -> Array:
    vmap_last = lambda fn: hk.vmap(fn, in_axes=-1, out_axes=-1, split_rng=False)
    tower_fn = vmap_last(vmap_last(self.net))
    if self.checkpoint_tower:
      tower_fn = hk.remat(tower_fn)
    return tower_fn(inputs)


@gin.register(denylist=['output_size'])
class Conv2DTower(hk.Module):
  """Two dimensional ConvNet tower module."""

  def __init__(
      self,
      output_size: int,
      num_hidden_units: int = gin.REQUIRED,
      num_hidden_layers: int = gin.REQUIRED,
      kernel_shape: Tuple[int, int] = gin.REQUIRED,
      with_bias: bool = True,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.relu,
      activate_final: bool = False,
      name: Optional[str] = None,
  ):
    """Tower that stacks up layers of ConvLonLat."""
    super().__init__(name=name)
    self.activation = activation
    self.activate_final = activate_final

    output_sizes = [num_hidden_units] * num_hidden_layers + [output_size]
    self.layers = []
    for output_size in output_sizes:
      self.layers.append(layers.ConvLonLat(
          output_size=output_size,
          kernel_shape=kernel_shape,
          with_bias=with_bias))

  def __call__(self, inputs: Array) -> Array:
    """Applies ConvNet tower to inputs."""
    num_layers = len(self.layers)
    out = inputs
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < (num_layers - 1) or self.activate_final:
        out = self.activation(out)
    return out


@gin.register(denylist=['output_size'])
class EpdTower(hk.Module):
  """EPD tower module parameterized by encode/process/decode factories."""

  def __init__(
      self,
      output_size: int,
      latent_size: int = gin.REQUIRED,
      num_process_blocks: int = gin.REQUIRED,
      encode_tower_factory: TowerFactory = gin.REQUIRED,
      process_tower_factory: TowerFactory = gin.REQUIRED,
      decode_tower_factory: TowerFactory = gin.REQUIRED,
      post_encode_activation: Optional[Callable[[Array], Array]] = None,
      pre_decode_activation: Optional[Callable[[Array], Array]] = None,
      final_activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.output_size = output_size
    self.latent_size = latent_size
    self.num_process_blocks = num_process_blocks
    self.encode_tower_factory = encode_tower_factory
    self.process_tower_factory = process_tower_factory
    self.decode_tower_factory = decode_tower_factory
    self.post_encode_activation = post_encode_activation
    self.pre_decode_activation = pre_decode_activation
    self.final_activation = final_activation

  def __call__(self, inputs: Array) -> Array:
    """Applies EPD tower to inputs."""
    encoded = self.encode_tower_factory(self.latent_size)(inputs)
    if self.post_encode_activation is not None:
      encoded = self.post_encode_activation(encoded)
    current = encoded
    for _ in range(self.num_process_blocks):
      current = current + self.process_tower_factory(self.latent_size)(current)
    if self.pre_decode_activation is not None:
      current = self.pre_decode_activation(current)
    out = self.decode_tower_factory(self.output_size)(current)
    if self.final_activation is not None:
      return self.final_activation(out)
    return out



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module-based API for calculating diagnostics of NeuralGCM models."""

import dataclasses

from flax import nnx
import jax.numpy as jnp

from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import data_specs
from neuralgcm.experimental import typing


class DiagnosticValue(nnx.Intermediate):
  """Variable type in which diagnostic values are stored."""

  ...


@dataclasses.dataclass
class DiagnosticModule(nnx.Module):
  """Base API for diagnostic modules."""

  def format_diagnostics(self, time: typing.Array) -> typing.Pytree:
    """Returns formatted diagnostics computed from the internal module state."""
    raise NotImplementedError(f'`format_diagnostics` on {self.__name__=}.')

  def __call__(self, *args, **kwargs) -> None:
    """Updates the internal module state from the inputs."""
    raise NotImplementedError(f'`__call__` on {self.__name__=}.')


# TODO(dkochkov) Generalize these to work on arbitrary pytrees.


class CumulativeDiagnostic(DiagnosticModule):
  """Diagnostic module that tracks cumulative value of an array stream."""

  def __init__(
      self,
      diagnostic_name: str,
      diagnostic_coords: cx.Coordinate,
      extract_fn=lambda x: x,
  ):
    self.diagnostic_name = diagnostic_name
    self.coords = diagnostic_coords
    self.extract_fn = extract_fn
    self.cumulative = DiagnosticValue(jnp.zeros(self.coords.shape))

  def format_diagnostics(self, time: typing.Array) -> typing.Pytree:
    return {
        self.diagnostic_name: data_specs.TimedField(
            cx.wrap(self.cumulative.value, self.coords), time
        )
    }

  def __call__(self, inputs):
    self.cumulative.value = self.cumulative.value + self.extract_fn(inputs)


class InstantDiagnostic(DiagnosticModule):
  """Diagnostic module that tracks instant value of an array stream."""

  def __init__(
      self,
      diagnostic_name: str,
      diagnostic_coords: cx.Coordinate,
      extract_fn=lambda x: x,
  ):
    self.diagnostic_name = diagnostic_name
    self.coords = diagnostic_coords
    self.extract_fn = extract_fn
    self.instant = DiagnosticValue(self.coords.shape)

  def format_diagnostics(self, time: typing.Array) -> typing.Pytree:
    return {
        self.diagnostic_name: data_specs.TimedField(
            cx.wrap(self.instant.value, self.coords), time
        )
    }

  def __call__(self, inputs):
    self.instant.value = self.extract_fn(inputs)


class IntervalDiagnostic(DiagnosticModule):
  """Diagnostic module that tracks interval cumulant of an array."""

  # TODO(dkochkov) verify this implementation and add tests.
  def __init__(
      self,
      diagnostic_name: str,
      interval_length: int,
      diagnostic_coords: cx.Coordinate,
      extract_fn=lambda x: x,
  ):
    self.diagnostic_name = diagnostic_name
    self.coords = diagnostic_coords
    self.interval_length = interval_length
    self.extract_fn = extract_fn
    self.cumulative = DiagnosticValue(jnp.zeros(self.coords.shape))
    interval_shape = (self.interval_length + 1,) + tuple(self.coords.shape)
    self.interval_values = DiagnosticValue(jnp.zeros(interval_shape))

  def next_interval(self, inputs):
    del inputs
    interval_values = self.interval_values.value
    c = self.cumulative.value
    interval_values = jnp.concat(
        [jnp.roll(interval_values, -1)[:-1], c[jnp.newaxis]]
    )
    self.interval_values.value = interval_values

  def format_diagnostics(self, time: typing.Array) -> typing.Pytree:
    interval_values = self.interval_values.value
    return {
        self.diagnostic_name: data_specs.TimedField(
            cx.wrap(self.cumulative.value - interval_values[0], self.coords),
            time,
        )
    }

  def __call__(self, inputs):
    self.cumulative.value = self.cumulative.value + self.extract_fn(inputs)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for processing and parsing gin configurations."""

import contextlib
import logging
import threading
import gin


_GIN_LOCK = threading.RLock()


def _remove_unknown_reference(gin_config_str: str) -> str:
  """Removes unknown references form `gin_config_str`."""
  # this happens when we have gin MACROS reference not imported objects.
  return '\n'.join([
      line for line in gin_config_str.splitlines()
      if 'gin.config._UnknownConfigurable' not in line
  ])


def parse_gin_config(
    physics_config_str: str,
    model_config_str: str,
    override_physics_configs_from_data: bool,
    gin_bindings: list[str],
):
  """Parses physics_config_str, model_config_str and gin_bindings in order.

  We use skip unknown parameters in model_config_str to avoid errors associated
  with irrelevant training parameters that refer to configurables only imported
  for training.

  Args:
    physics_config_str: gin configuration string for physics_specifications
      object that stores relevant physics constants.
    model_config_str: gin configuration string of the model.
    override_physics_configs_from_data: whether to reparse `physics_config_str`
      after processing `model_config_str`.
    gin_bindings: additional gin configuration strings that will be parsed last.
  """
  gin.parse_config(physics_config_str)
  gin.parse_config(model_config_str, skip_unknown=True)
  if override_physics_configs_from_data:
    gin.parse_config(physics_config_str)
  gin.parse_config(gin_bindings)
  logging.info('Evaluating model with the following config:\n %s',
               gin.config_str())


@contextlib.contextmanager
def specific_config(
    gin_config: str,
    clear_current: bool = True,
    skip_unknown: bool = True,
):
  """Context manager for evaluation of functions with `gin_config`."""
  with _GIN_LOCK:
    # avoid splitting long lines into multiples that may contain unknown refs.
    current_config = gin.config_str(max_line_length=len(gin.config_str()))
    current_config = _remove_unknown_reference(current_config)
    if clear_current:
      gin.clear_config()
    try:
      gin.parse_config(gin_config, skip_unknown=skip_unknown)
      yield
    finally:
      gin.clear_config()
      gin.parse_config(current_config)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for Metrics."""
import dataclasses
from typing import Callable
from dinosaur import typing
import jax
import jax.numpy as jnp
import metrics_util


Pytree = typing.Pytree
TrajectoryRepresentations = typing.TrajectoryRepresentations

tree_leaves = jax.tree_util.tree_leaves
tree_map = jax.tree_util.tree_map


@dataclasses.dataclass
class Evaluator:
  """Class that evaluates on (prediction, trajectory) returning Pytree."""

  def evaluate(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> Pytree:
    """Evaluates giving values of interest."""
    raise NotImplementedError()


@dataclasses.dataclass
class EvaluateFunctionWrapper(Evaluator):
  """Wraps `evaluate_fn` function to be used as an Evaluator."""

  def __init__(
      self,
      evaluate_fn: Callable[
          [TrajectoryRepresentations, TrajectoryRepresentations], Pytree
      ],
  ):
    self._evaluate_fn = evaluate_fn

  def evaluate(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> Pytree:
    return self._evaluate_fn(prediction, target)


class MetricRuntimeError(Exception):
  """Generic error for Metrics to raise in place of generic RuntimeError."""


@dataclasses.dataclass
class Metric(Evaluator):
  """An Evaluator that derives information from a TrajectorySpec."""

  trajectory_spec: metrics_util.TrajectorySpec
  is_nodal: bool = dataclasses.field(default=True, kw_only=True)
  is_encoded: bool = dataclasses.field(default=False, kw_only=True)

  def get_representation(self, x: TrajectoryRepresentations) -> Pytree:
    x_rep = x.get_representation(
        is_nodal=self.is_nodal, is_encoded=self.is_encoded
    )
    if x_rep is None:
      raise MetricRuntimeError(
          'Desired representation of `x` was None. '
          f'{self.is_nodal=}, {self.is_encoded=}'
      )
    return x_rep

  def surface_mean(self, trajectory: Pytree) -> Pytree:
    if self.is_encoded:
      coords = self.trajectory_spec.coords
    else:
      coords = self.trajectory_spec.data_coords
    if self.is_nodal:
      # Mean over lat/lon. Converts shapes
      #   (n_time, n_level, n_lon, n_lat) --> (n_time, n_level)
      fn = lambda x: metrics_util.nodal_surface_mean(x, coords)
    else:
      fn = lambda x: metrics_util.modal_surface_mean(x, coords)
    return tree_map(fn, trajectory)

  def mean_per_variable(self, trajectory: Pytree) -> Pytree:
    # In practice this is used to reduce shape (n_time, n_level) --> ()
    return tree_map(jnp.mean, self.surface_mean(trajectory))


class ScalarMetric(Metric):
  """Metric that compute scalar quantities."""


@dataclasses.dataclass
class Loss(ScalarMetric):
  """Metric that can be used as a loss."""

  trajectory_spec: metrics_util.TrajectorySpec
  is_nodal: bool = dataclasses.field(default=True, kw_only=True)
  is_encoded: bool = dataclasses.field(default=False, kw_only=True)
  time_step: int | slice | None = dataclasses.field(default=None, kw_only=True)

  def evaluate_per_variable(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> Pytree:
    raise NotImplementedError()

  def evaluate(
      self,
      prediction: TrajectoryRepresentations,
      target: TrajectoryRepresentations,
  ) -> jnp.ndarray:
    error_per_variable = self.evaluate_per_variable(prediction, target)
    return sum(tree_leaves(error_per_variable))

  def debug_loss_terms_instance(self) -> EvaluateFunctionWrapper:
    """Returns class that evaluates relative loss per variable."""

    def evaluate_fn(
        prediction: TrajectoryRepresentations,
        target: TrajectoryRepresentations,
    ) -> Pytree:
      # self.loss.evaluate takes ensemble mean (to evaluate on ensemble mean) if
      # needed.
      loss_per_variable = self.evaluate_per_variable(prediction, target)
      # here we reduce terms by summation to expose relative contributions,
      # even though the actual total_loss might be different.
      sum_of_all_terms = sum(tree_leaves(loss_per_variable))
      relative_loss = tree_map(
          lambda x: x / sum_of_all_terms, loss_per_variable
      )
      return {'relative_loss': relative_loss}

    return EvaluateFunctionWrapper(evaluate_fn)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for forcings.py."""

import functools
from unittest import mock
from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import xarray_utils
import haiku as hk
import jax
from neuralgcm import forcings
import numpy as np


units = scales.units


class ForcingsTest(parameterized.TestCase):
  """Tests corrector modules."""

  def setUp(self):
    super().setUp()
    n_sigma_layers = 3
    self.coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(n_sigma_layers)
    )
    self.physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = self.physics_specs.nondimensionalize(10 * units.minute)
    ref_datetime = np.datetime64('1970-01-01T00:00:00')
    self.aux_features = {xarray_utils.REFERENCE_DATETIME_KEY: ref_datetime}

  def test_no_forcing(self):

    def forcing_fwd(forcing_data, sim_time):
      forcing_fn = forcings.NoForcing(self.coords, self.dt, self.physics_specs,
                                      self.aux_features)
      return forcing_fn(forcing_data, sim_time)

    forcing_model = hk.without_apply_rng(hk.transform(forcing_fwd))
    forcing_data = {}
    sim_time = None
    params = forcing_model.init(jax.random.PRNGKey(42), forcing_data, sim_time)
    forcing = forcing_model.apply(params, forcing_data, sim_time)

    expected_forcing = {}
    self.assertEqual(forcing, expected_forcing)

  def test_check_errors(self):
    # The mock in test_dynamic_data_forcing checks that
    # _FORCING_ERRORS.append was indeed called the correct number of times.

    with self.assertRaisesRegex(forcings.ForcingDataError, 'forcing_sim_time'):
      forcings._check_errors(err_list=['forcing_sim_time was bad'])

  @mock.patch('neuralgcm.forcings._FORCING_ERRORS', new=mock.MagicMock(list))
  def test_dynamic_data_forcing(self):
    forcings._check_errors()  # Should not raise

    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    one_hour_nondim = physics_specs.nondimensionalize(units.hour)
    n_times = 120
    sim_time_data = one_hour_nondim * np.arange(n_times)
    surface_shape = (n_times,) + self.coords.surface_nodal_shape
    volume_shape = (n_times,) + self.coords.nodal_shape
    # specify time-dependent data
    snow_temperature = 260 * np.ones(surface_shape)
    snow_depth = np.arange(n_times).reshape((n_times, 1, 1, 1)) * np.ones(
        surface_shape)
    cloud_cover = np.zeros(volume_shape)
    forcing_data = {'cloud_cover': cloud_cover,
                    'snow_temperature': snow_temperature,
                    'snow_depth': snow_depth,
                    'sim_time': sim_time_data}
    inputs_to_units_mapping = {'cloud_cover': 'dimensionless',
                               'snow_temperature': 'K',
                               'snow_depth': 'meter',
                               'sim_time': 'dimensionless'}

    def forcing_fwd(forcing_data, sim_time):
      forcing_fn = forcings.DynamicDataForcing(
          self.coords, self.dt, self.physics_specs, self.aux_features,
          check_sim_time_errors=True,
          inputs_to_units_mapping=inputs_to_units_mapping)
      return forcing_fn(forcing_data, sim_time)

    forcing_model = hk.without_apply_rng(hk.transform(forcing_fwd))
    params = forcing_model.init(jax.random.PRNGKey(42), forcing_data, 0)
    forcing_fn = jax.jit(forcing_model.apply)

    with self.subTest('forcing shapes'):
      sim_time = 0
      forcing = forcing_fn(params, forcing_data, sim_time)
      self.assertSetEqual(
          set(forcing.keys()),
          {'cloud_cover', 'snow_temperature', 'snow_depth', 'sim_time'})
      self.assertEqual(forcing['cloud_cover'].shape, (3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, ())

    with self.subTest('exact sim_time'):
      for idx, sim_time in enumerate(sim_time_data):
        forcing = forcing_fn(params, forcing_data, sim_time)
        self.assertFalse(np.any(np.isnan(forcing['sim_time'])))
        np.testing.assert_allclose(forcing['sim_time'], sim_time)
        np.testing.assert_allclose(forcing['snow_temperature'], 260)
        nondim_snow_depth = self.physics_specs.nondimensionalize(
            idx * units.meter)
        np.testing.assert_allclose(forcing['snow_depth'],
                                   nondim_snow_depth, atol=1e-6)

    with self.subTest('nearby sim_time'):
      for idx, sim_time in enumerate(sim_time_data):
        nearby_sim_time = sim_time + one_hour_nondim / 12  # +5 minutes
        forcing = forcing_fn(params, forcing_data, nearby_sim_time)
        self.assertFalse(np.any(np.isnan(forcing['sim_time'])))
        np.testing.assert_allclose(forcing['sim_time'], sim_time)
        nondim_snow_depth = self.physics_specs.nondimensionalize(
            idx * units.meter)
        np.testing.assert_allclose(forcing['snow_depth'],
                                   nondim_snow_depth, atol=1e-6)

    with self.subTest('in-between sim_time'):
      sim_time = 3.5 * one_hour_nondim
      forcing = forcing_fn(params, forcing_data, sim_time)
      np.testing.assert_allclose(forcing['sim_time'], sim_time_data[4])
      nondim_snow_depth = self.physics_specs.nondimensionalize(
          forcing_data['snow_depth'][4] * units.meter)
      np.testing.assert_allclose(forcing['snow_depth'],
                                 nondim_snow_depth, atol=1e-6)

    with self.subTest('thinned sim_time'):
      for idx, sim_time in list(enumerate(sim_time_data))[::10]:
        thinned_forcing_data = {k: v[::10] for k, v in forcing_data.items()}
        forcing = forcing_fn(params, thinned_forcing_data, sim_time)
        self.assertFalse(np.any(np.isnan(forcing['sim_time'])))
        np.testing.assert_allclose(forcing['sim_time'], sim_time)
        np.testing.assert_allclose(forcing['snow_temperature'], 260)
        nondim_snow_depth = self.physics_specs.nondimensionalize(
            idx * units.meter)
        np.testing.assert_allclose(forcing['snow_depth'],
                                   nondim_snow_depth, atol=1e-6)

    with self.subTest('sim_time too small'):
      sim_time = -1 * one_hour_nondim
      jax.block_until_ready(forcing_fn(params, forcing_data, sim_time))
      self.assertEqual(1, forcings._FORCING_ERRORS.append.call_count)

    with self.subTest('sim_time too big'):
      sim_time = (n_times + 1) * one_hour_nondim
      jax.block_until_ready(forcing_fn(params, forcing_data, sim_time))
      self.assertEqual(2, forcings._FORCING_ERRORS.append.call_count)

    with self.subTest('sim_time is None'):
      sim_time = None
      with self.assertRaises(TypeError):
        forcing_fn(params, forcing_data, sim_time)

    with self.subTest('sim_time is nan'):
      # jax casts None to an array with nan
      sim_time = jax.numpy.asarray(None)  # = DeviceArray(nan, dtype=float32)
      # inside forcing_fn, sim_time gets cast to int32
      # this is unsafe, since it turns nan into an actual integer,
      # idx=2147483647.
      jax.block_until_ready(forcing_fn(params, forcing_data, sim_time))
      self.assertEqual(3, forcings._FORCING_ERRORS.append.call_count)

    with self.subTest('vmap sim_time'):
      sim_time = sim_time_data[2:4]

      def forcing_fwd_vmap(forcing_data, sim_time):
        forcing_fn = forcings.DynamicDataForcing(
            self.coords, self.dt, self.physics_specs, self.aux_features,
            check_sim_time_errors=True,
            inputs_to_units_mapping=inputs_to_units_mapping)
        forcing_fn = jax.vmap(forcing_fn, in_axes=(None, 0))
        return forcing_fn(forcing_data, sim_time)

      forcing_vmap_model = hk.without_apply_rng(hk.transform(forcing_fwd_vmap))
      params = forcing_vmap_model.init(
          jax.random.PRNGKey(42), forcing_data, sim_time)

      forcing = forcing_vmap_model.apply(params, forcing_data, sim_time)
      self.assertSetEqual(
          set(forcing.keys()),
          {'cloud_cover', 'snow_temperature', 'snow_depth', 'sim_time'})
      self.assertEqual(forcing['cloud_cover'].shape, (2, 3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (2, 1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (2, 1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, (2,))
      np.testing.assert_allclose(forcing['sim_time'], sim_time_data[2:4])

    with self.subTest('single sim_time forcing_data'):
      single_forcing_data = {k: v[2:3] for k, v in forcing_data.items()}
      forcing = forcing_fn(params, single_forcing_data, sim_time_data[2])
      self.assertEqual(forcing['cloud_cover'].shape, (3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, ())
      np.testing.assert_allclose(forcing['sim_time'], sim_time_data[2])

  def test_persistence_data_forcing(self):
    sim_time_data = 1.0 * np.arange(5)
    surface_shape = (len(sim_time_data),) + self.coords.surface_nodal_shape
    volume_shape = (len(sim_time_data),) + self.coords.nodal_shape
    # specify time-dependent data
    snow_temperature = 260 * np.ones(surface_shape)
    snow_depth = np.arange(5).reshape((5, 1, 1, 1)) * np.ones(surface_shape)
    cloud_cover = np.zeros(volume_shape)
    forcing_data = {'cloud_cover': cloud_cover,
                    'snow_temperature': snow_temperature,
                    'snow_depth': snow_depth,
                    'sim_time': sim_time_data}
    inputs_to_units_mapping = {'cloud_cover': 'dimensionless',
                               'snow_temperature': 'K',
                               'snow_depth': 'meter',
                               'sim_time': 'dimensionless'}

    def forcing_fwd(forcing_data, sim_time):
      forcing_fn = forcings.PersistenceDataForcing(
          self.coords, self.dt, self.physics_specs, self.aux_features,
          inputs_to_units_mapping=inputs_to_units_mapping)
      return forcing_fn(forcing_data, sim_time)

    forcing_model = hk.without_apply_rng(hk.transform(forcing_fwd))
    params = forcing_model.init(jax.random.PRNGKey(42), forcing_data, 0)
    forcing_fn = jax.jit(forcing_model.apply)

    with self.subTest('forcing shapes'):
      sim_time = 0
      forcing = forcing_fn(params, forcing_data, sim_time)
      self.assertSetEqual(
          set(forcing.keys()),
          {'cloud_cover', 'snow_temperature', 'snow_depth', 'sim_time'})
      self.assertEqual(forcing['cloud_cover'].shape, (3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, ())

    with self.subTest('numeric sim_time'):
      for sim_time in [-1.0, 0, 1.2, 3.5]:
        forcing = forcing_fn(params, forcing_data, sim_time)
        expected_forcing = forcing_fn(params, forcing_data, sim_time_data[0])
        for v, v_expected in zip(forcing.values(), expected_forcing.values()):
          np.testing.assert_allclose(v, v_expected)

    with self.subTest('sim_time is None'):
      sim_time = None
      forcing = forcing_fn(params, forcing_data, sim_time)
      np.testing.assert_allclose(forcing['sim_time'], sim_time_data[0])

    with self.subTest('sim_time is nan'):
      # jax casts None to an array with nan
      sim_time = jax.numpy.asarray(None)  # = DeviceArray(nan, dtype=float32)
      forcing = forcing_fn(params, forcing_data, sim_time)
      np.testing.assert_allclose(forcing['sim_time'], sim_time_data[0])

    with self.subTest('vmap sim_time'):
      sim_time = np.array([2., 3.])

      def forcing_fwd_vmap(forcing_data, sim_time):
        forcing_fn = forcings.PersistenceDataForcing(
            self.coords, self.dt, self.physics_specs, self.aux_features,
            inputs_to_units_mapping=inputs_to_units_mapping)
        forcing_fn = jax.vmap(forcing_fn, in_axes=(None, 0))
        return forcing_fn(forcing_data, sim_time)

      forcing_vmap_model = hk.without_apply_rng(hk.transform(forcing_fwd_vmap))
      params = forcing_vmap_model.init(
          jax.random.PRNGKey(42), forcing_data, sim_time)

      forcing = forcing_vmap_model.apply(params, forcing_data, sim_time)
      self.assertSetEqual(
          set(forcing.keys()),
          {'cloud_cover', 'snow_temperature', 'snow_depth', 'sim_time'})
      self.assertEqual(forcing['cloud_cover'].shape, (2, 3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (2, 1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (2, 1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, (2,))
      np.testing.assert_allclose(forcing['sim_time'], [0., 0.])

    with self.subTest('single sim_time forcing_data'):
      single_forcing_data = {k: v[2:3] for k, v in forcing_data.items()}
      forcing = forcing_fn(params, single_forcing_data, sim_time=4.0)
      self.assertEqual(forcing['cloud_cover'].shape, (3, 64, 32))
      self.assertEqual(forcing['snow_temperature'].shape, (1, 64, 32))
      self.assertEqual(forcing['snow_depth'].shape, (1, 64, 32))
      self.assertEqual(forcing['sim_time'].shape, ())
      np.testing.assert_allclose(forcing['sim_time'], 2.)  # 0 index value

  @parameterized.parameters(
      dict(forcing_module=forcings.DynamicDataForcing),
      dict(forcing_module=forcings.PersistenceDataForcing),
  )
  def test_forcing_transforms(self, forcing_module):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    one_hour_nondim = physics_specs.nondimensionalize(units.hour)
    n_times = 120
    sim_time_data = one_hour_nondim * np.arange(n_times)
    surface_shape = (n_times,) + self.coords.surface_nodal_shape
    # specify time-dependent data
    sea_surface_temperature = 334 * np.ones(surface_shape)

    forcing_data = {'sea_surface_temperature': sea_surface_temperature,
                    'sim_time': sim_time_data}
    inputs_to_units_mapping = {'sea_surface_temperature': 'K',
                               'sim_time': 'dimensionless'}

    with self.subTest('IncrementSSTForcingTransform'):
      forcing_transform = functools.partial(
          forcings.IncrementSSTForcingTransform,
          temperature_change='2.5 K',
          key='sea_surface_temperature',
      )

      def forcing_fwd(forcing_data, sim_time):
        forcing_fn = forcing_module(
            self.coords, self.dt, self.physics_specs, self.aux_features,
            inputs_to_units_mapping=inputs_to_units_mapping,
            forcing_transform=forcing_transform)
        return forcing_fn(forcing_data, sim_time)

      forcing_model = hk.without_apply_rng(hk.transform(forcing_fwd))
      params = forcing_model.init(jax.random.PRNGKey(42), forcing_data, 0)
      forcing_fn = jax.jit(forcing_model.apply)

      sim_time = 0
      forcing = forcing_fn(params, forcing_data, sim_time)
      self.assertEqual(forcing['sea_surface_temperature'].shape, (1, 64, 32))
      np.testing.assert_allclose(forcing['sea_surface_temperature'], 334 + 2.5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for features.py."""

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
from neuralgcm import embeddings  # pylint: disable=unused-import
from neuralgcm import features  # pylint: disable=unused-import
from neuralgcm import mappings  # pylint: disable=unused-import
from neuralgcm import physics_specifications
import numpy as np


units = scales.units


class FeaturesTest(parameterized.TestCase):
  """Tests feature modules."""

  @parameterized.parameters(
      ('VelocityAndPrognostics', 'modal'),
      ('NodalInputVelocityAndPrognostics', 'nodal'),
      ('RadiationFeatures', 'modal'),
      ('RadiationFeatures', 'nodal'),
      ('OrbitalTimeFeatures', 'modal'),
      ('OrbitalTimeFeatures', 'nodal'),
      ('ForcingFeatures', 'modal'),
      ('ForcingFeatures', 'nodal'),
      ('LatitudeFeatures', 'modal'),
      ('LatitudeFeatures', 'nodal'),
      ('OrographyFeatures', 'modal'),
      ('OrographyFeatures', 'nodal'),
      ('OneHotAuxFeatures', 'modal'),
      ('OneHotAuxFeatures', 'nodal'),
      ('LearnedPositionalFeatures', 'modal'),
      ('LearnedPositionalFeatures', 'nodal'),
      ('CombinedFeatures', 'modal'),
      ('CombinedFeatures', 'nodal'),
      ('MemoryVelocityAndValues', 'modal'),
      ('RandomnessFeatures', 'modal'),
      ('RandomnessFeatures', 'modal', True),
  )
  def test_feature_module_shapes(
      self,
      features_name,
      input_type,
      dict_randomness=False,
  ):
    """Tests that the returned features have expected nodal shape."""
    gin.enter_interactive_mode()
    gin.clear_config()
    physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @primitive_eq_specs_constructor',
    ])
    gin.parse_config(physics_gin_config)
    physics_specs = physics_specifications.get_physics_specs()
    dt = physics_specs.nondimensionalize(100 * units.s)
    # setting up coordinates and initial state.
    grid = spherical_harmonic.Grid.T21()
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(4)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs
    )
    # handling float and integer covariates
    fake_land_sea_mask = np.random.random(grid.nodal_shape)  # float [0, 1)
    aux_features['land_sea_mask'] = fake_land_sea_mask
    fake_soil_type = np.random.randint(0, 8, grid.nodal_shape)
    aux_features['soil_type'] = fake_soil_type
    reference_datetime = np.datetime64('1970-01-01T00:00:00')
    aux_features[xarray_utils.REFERENCE_DATETIME_KEY] = reference_datetime
    inputs = initial_state_fn(jax.random.PRNGKey(42)).asdict()
    inputs['tracers'] = {'specific_humidity': inputs['divergence']}
    inputs['sim_time'] = 0.5
    if input_type == 'nodal':
      inputs = coords.horizontal.to_nodal(inputs)
    forcing = {
        'total_cloud_cover': np.random.random(coords.surface_nodal_shape),
        'sim_time': 0.5,
    }

    @gin.configurable
    def compute_feature_fwd(
        x, memory, diagnostics, randomness, forcing, feature_module
    ):
      feature_fn = feature_module(coords, dt, physics_specs, aux_features)
      return feature_fn(x, memory, diagnostics, randomness, forcing)

    gin_config = '\n'.join([
        f'compute_feature_fwd.feature_module = @{features_name}',
        'OneHotAuxFeatures.covariate_keys = ("land_sea_mask", "soil_type")',
        'OneHotAuxFeatures.convert_float_to_int = True',
        'LearnedPositionalFeatures.latent_size = 4',
        'ForcingFeatures.forcing_to_include = ("total_cloud_cover",)',
        'CombinedFeatures.feature_modules ='
        + ' (@OneHotAuxFeatures, @RadiationFeatures)',
    ])
    gin.parse_config(gin_config)
    model = hk.without_apply_rng(hk.transform(compute_feature_fwd))
    rng = jax.random.PRNGKey(42)
    # in this test we assume that `memory` has the same structure as state.
    memory = inputs
    diagnostics = None
    if dict_randomness:
      randomness = {
          f'random_field_{i:01}': jax.random.uniform(rng, grid.nodal_shape)
          for i in range(3)
      }
    else:
      randomness = jax.random.uniform(rng, grid.nodal_shape)
    params = model.init(rng, inputs, memory, diagnostics, randomness, forcing)
    feature_values = model.apply(
        params, inputs, memory, diagnostics, randomness, forcing
    )
    for _, v in feature_values.items():
      self.assertEqual(v.shape[-2:], grid.nodal_shape)
      self.assertEqual(v.ndim, 3)


@gin.register
class TowerMock(hk.Module):
  """Tower that returns random outputs with shape (output_size,) + nodal_shape."""

  def __init__(
      self,
      output_size: int,  # size of dim 0
      nodal_shape: tuple[int, ...] = gin.REQUIRED,  # sizes of remaining dims
      **unused_kwargs,
  ):
    super().__init__(name=None)
    self.output_size = output_size
    self.nodal_shape = nodal_shape

  def __call__(self, inputs):
    del inputs  # unused.
    return jax.random.uniform(
        jax.random.PRNGKey(42), shape=(self.output_size,) + self.nodal_shape
    )


class EmbeddingFeaturesTest(parameterized.TestCase):
  """Tests embedding feature modules."""

  @parameterized.parameters(
      dict(
          features_name='EmbeddingSurfaceFeatures',
          embedding_name='ModalToNodalEmbedding',
          mapping_name='NodalMapping',
          tower_nodal_shape=(64, 32),
          input_type='modal',
      ),
      dict(
          features_name='EmbeddingVolumeFeatures',
          embedding_name='ModalToNodalEmbedding',
          mapping_name='NodalVolumeMapping',
          tower_nodal_shape=(4, 64, 32),
          input_type='modal',
      ),
  )
  def test_embedding_feature_module_shapes(
      self,
      features_name,
      embedding_name,
      mapping_name,
      tower_nodal_shape,
      input_type,
  ):
    """Tests that the returned features have expected nodal shape."""
    gin.enter_interactive_mode()
    gin.clear_config()
    physics_gin_config = '\n'.join([
        'get_physics_specs.construct_fn = @primitive_eq_specs_constructor',
    ])
    gin.parse_config(physics_gin_config)
    physics_specs = physics_specifications.get_physics_specs()
    dt = physics_specs.nondimensionalize(100 * units.s)
    # setting up coordinates and initial state.
    grid = spherical_harmonic.Grid.T21()
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(4)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs
    )
    inputs = initial_state_fn(jax.random.PRNGKey(42)).asdict()
    inputs['sim_time'] = 0.5
    if input_type == 'nodal':
      inputs = coords.horizontal.to_nodal(inputs)
    forcing = {}

    @gin.configurable
    def compute_feature_fwd(
        x,
        memory,
        diagnostics,
        randomness,
        forcing,
        feature_module,
    ):
      feature_fn = feature_module(coords, dt, physics_specs, aux_features)
      return feature_fn(x, memory, diagnostics, randomness, forcing)

    gin_config = '\n'.join([
        f'compute_feature_fwd.feature_module = @{features_name}',
        # Config for EmbeddingSurfaceFeatures
        f'{features_name}.feature_name = "embedding_features"',
        f'{features_name}.output_size = 8',
        f'{features_name}.embedding_module = @{embedding_name}',
        f'{embedding_name}.modal_to_nodal_features_module ='
        + ' @VelocityAndPrognostics',
        'VelocityAndPrognostics.fields_to_include = ("u", "v")',
        f'{embedding_name}.nodal_mapping_module = @{mapping_name}',
        f'{mapping_name}.tower_factory = @TowerMock',
        f'TowerMock.nodal_shape = {tower_nodal_shape}',
    ])
    gin.parse_config(gin_config)
    model = hk.without_apply_rng(hk.transform(compute_feature_fwd))
    rng = jax.random.PRNGKey(42)
    # in this test we assume that `memory` has the same structure as state.
    memory = inputs
    diagnostics = None
    randomness = None
    params = model.init(rng, inputs, memory, diagnostics, randomness, forcing)
    feature_values = model.apply(
        params, inputs, memory, diagnostics, randomness, forcing
    )
    for _, v in feature_values.items():
      self.assertEqual(v.shape[-2:], grid.nodal_shape)
      self.assertEqual(v.ndim, 3)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for mappings.py."""

from absl.testing import absltest
from absl.testing import parameterized
import gin
import haiku as hk
import jax
import jax.numpy as jnp

from neuralgcm import mappings  # pylint: disable=unused-import
from neuralgcm import transforms  # pylint: disable=unused-import
import numpy as np


@gin.register
class MappingsTowerMock(hk.Module):
  """Tower that returns random outputs with same nodal shape and ndim as inputs."""

  def __init__(self, output_size: int):
    super().__init__(name=None)
    self.output_size = output_size

  def __call__(self, inputs, latents=None, position_encoding=None):
    nodal_shape = inputs.shape[1:]
    return jax.random.uniform(
        jax.random.PRNGKey(42), shape=(self.output_size,) + nodal_shape
    )


class MappingsTest(parameterized.TestCase):
  """Tests mappings modules."""

  # Sharp bits: output shape values need to be arrays, otherwise mapping over
  # pytree leaves will return the integers of the shape instead.
  # Using `input_shapes` to set inputs to make sponge link more readable
  @parameterized.parameters(
      dict(
          mapping_name='NodalMapping',
          input_shapes={
              'u': np.asarray((13, 64, 32)),
              'v': np.asarray((13, 64, 32)),
              'cos_latitude': np.asarray((1, 64, 32)),
              'learned_feature': np.asarray((4, 64, 32)),
          },
          # the tower output_size will be 5 (sum of dim=0 sizes)
          output_shapes={
              'prediction_1': np.asarray((4, 64, 32)),
              'prediction_2': np.asarray((4, 64, 32)),
              'nesting_key': {
                  'nested_prediction': np.asarray((1, 64, 32)),
              },
          },
          value_error=None,
      ),
      dict(
          mapping_name='NodalVolumeMapping',
          input_shapes={
              'u': np.asarray((13, 64, 32)),
              'v': np.asarray((13, 64, 32)),
          },
          # the tower output_size will be 3 (number of keys)
          output_shapes={
              'prediction_1': np.asarray((13, 64, 32)),
              'prediction_2': np.asarray((13, 64, 32)),
              'nesting_key': {
                  'nested_prediction_1': np.asarray((13, 64, 32)),
                  'nested_prediction_2': np.asarray((13, 64, 32)),
              },
          },
          value_error=None,
      ),
      dict(
          mapping_name='NodalVolumeMapping',
          input_shapes={
              'u': np.asarray((1, 64, 32)),
              'v': np.asarray((4, 64, 32)),  # can't stack with u
          },
          # the tower output_size will be 1 (number of keys)
          output_shapes={'prediction': np.asarray((1, 64, 32))},
          value_error='All input arrays must have the same shape.',
      ),
      dict(
          mapping_name='NodalVolumeTransformerMapping',
          input_shapes={
              'u': np.asarray((1, 64, 32)),
              'v': np.asarray((4, 64, 32)),  # can't stack with u
          },
          # the tower output_size will be 1 (number of keys)
          output_shapes={'prediction': np.asarray((1, 64, 32))},
          value_error='All input arrays must have the same shape.',
      ),
  )
  def test_nodal_mapping_shapes(
      self, mapping_name, input_shapes, output_shapes, value_error,
  ):
    """Tests that mapping produces expected output structures."""
    gin.enter_interactive_mode()
    gin.clear_config()

    @gin.configurable
    def mapping_forward(x, mapping):  # defines a forward function
      return mapping(output_shapes)(x)

    long_name = 'NodalVolumeTransformerMapping'
    config = '\n'.join([
        f'mapping_forward.mapping = @{mapping_name}',
        'NodalMapping.tower_factory = @MappingsTowerMock',
        'NodalVolumeMapping.tower_factory = @MappingsTowerMock',
        f'{long_name}.latent_size = 5',
        f'{long_name}.encoder_transformer_tower_factory = @MappingsTowerMock',
        f'{long_name}.decoder_transformer_tower_factory = @MappingsTowerMock',
        f'{long_name}.encoder_inputs_selection_module = @FeatureSelector',
        'FeatureSelector.regex_patterns = ".*"'  # selects all features.
    ])
    gin.parse_config(config)

    inputs = jax.tree_util.tree_map(jnp.ones, input_shapes)
    model = hk.without_apply_rng(hk.transform(mapping_forward))
    if value_error is not None:
      with self.assertRaisesRegex(ValueError, value_error):
        params = model.init(jax.random.PRNGKey(42), inputs)
    else:
      params = model.init(jax.random.PRNGKey(42), inputs)
      output = model.apply(params, inputs)
      for output_item, expected_shape in zip(
          jax.tree_util.tree_leaves(output),
          jax.tree_util.tree_leaves(output_shapes)):
        self.assertEqual(output_item.shape, tuple(expected_shape.tolist()))


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests that pytree mappings produce outputs with expected shapes."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
import chex
from flax import nnx
import jax
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import pytree_mappings
from neuralgcm.experimental import pytree_transforms
from neuralgcm.experimental import pytree_utils
from neuralgcm.experimental import standard_layers
from neuralgcm.experimental import towers
from neuralgcm.experimental import typing
import numpy as np


class EmbeddingsTest(parameterized.TestCase):
  """Tests embedding modules."""

  def _test_embedding_module(
      self,
      embedding_module: nnx.Module,
      inputs: typing.Pytree,
  ):
    embedded_features = embedding_module(inputs)
    actual = embedding_module.output_shapes
    expected = pytree_utils.shape_structure(embedded_features)
    chex.assert_trees_all_equal(actual, expected)

  def test_embedding(self):
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.SigmaLevels.equidistant(4),
    )
    input_names = ('u', 'v')
    feature_module = pytree_transforms.PrognosticFeatures(
        coords, volume_field_names=input_names
    )
    tower_factory = functools.partial(
        towers.ColumnTower,
        column_net_factory=functools.partial(
            standard_layers.MlpUniform, hidden_size=6, n_hidden_layers=2
        ),
    )
    mapping_factory = functools.partial(
        pytree_mappings.ChannelMapping,
        tower_factory=tower_factory,
    )
    output_shapes = {
        'a': typing.ShapeFloatStruct((7,) + coords.horizontal.shape),
        'b': typing.ShapeFloatStruct((3,) + coords.horizontal.shape),
    }
    embedding = pytree_mappings.Embedding(
        output_shapes=output_shapes,
        feature_module=feature_module,
        mapping_factory=mapping_factory,
        rngs=nnx.Rngs(0),
    )
    test_inputs = {k: np.ones(coords.shape) for k in input_names}
    self._test_embedding_module(embedding, test_inputs)

  def test_coordinate_state_mapping(self):
    """Checks that CoordsStateMapping produces outputs with expected shapes."""
    coords = coordinates.DinosaurCoordinates(
        horizontal=coordinates.LonLatGrid.T21(),
        vertical=coordinates.SigmaLevels.equidistant(4),
    )
    input_names = ('u', 'v')
    feature_module = pytree_transforms.PrognosticFeatures(
        coords, volume_field_names=input_names
    )
    tower_factory = functools.partial(
        towers.ColumnTower,
        column_net_factory=functools.partial(
            standard_layers.MlpUniform, hidden_size=6, n_hidden_layers=2
        ),
    )
    mapping_factory = functools.partial(
        pytree_mappings.ChannelMapping,
        tower_factory=tower_factory,
    )
    embedding_factory = functools.partial(
        pytree_mappings.Embedding,
        feature_module=feature_module,
        mapping_factory=mapping_factory,
    )
    volume_field_names = ('u', 'div')
    surface_field_names = ('pressure',)
    state_mapping = pytree_mappings.CoordsStateMapping(
        coords=coords,
        surface_field_names=surface_field_names,
        volume_field_names=volume_field_names,
        embedding_factory=embedding_factory,
        rngs=nnx.Rngs(0),
    )
    test_inputs = {k: np.ones(coords.shape) for k in input_names}
    out = state_mapping(test_inputs)
    out_shape = pytree_utils.shape_structure(out)
    expected_shape = {
        'u': typing.ShapeFloatStruct(coords.shape),
        'div': typing.ShapeFloatStruct(coords.shape),
        'pressure': typing.ShapeFloatStruct((1,) + coords.horizontal.shape),
    }
    chex.assert_trees_all_equal(out_shape, expected_shape)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import operator
from typing import Any, Callable

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
from neuralgcm.experimental import coordax
from neuralgcm.experimental.coordax import core
from neuralgcm.experimental.coordax import testing
import numpy as np


class CoreTest(parameterized.TestCase):

  PRODUCT_XY = coordax.CartesianProduct(
      (coordax.NamedAxis('x', 2), coordax.NamedAxis('y', 3))
  )

  @parameterized.named_parameters(
      dict(
          testcase_name='empty',
          coordinates=(),
          expected=(),
      ),
      dict(
          testcase_name='single_other_axis',
          coordinates=(coordax.NamedAxis('x', 2),),
          expected=(coordax.NamedAxis('x', 2),),
      ),
      dict(
          testcase_name='single_selected_axis',
          coordinates=(
              coordax.SelectedAxis(coordax.NamedAxis('x', 2), axis=0),
          ),
          expected=(coordax.NamedAxis('x', 2),),
      ),
      dict(
          testcase_name='pair_of_other_axes',
          coordinates=(
              coordax.NamedAxis('x', 2),
              coordax.LabeledAxis('y', np.arange(3)),
          ),
          expected=(
              coordax.NamedAxis('x', 2),
              coordax.LabeledAxis('y', np.arange(3)),
          ),
      ),
      dict(
          testcase_name='pair_of_selections_correct',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
          ),
          expected=(PRODUCT_XY,),
      ),
      dict(
          testcase_name='pair_of_selections_wrong_order',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
          ),
          expected=(
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
          ),
      ),
      dict(
          testcase_name='selection_incomplete',
          coordinates=(coordax.SelectedAxis(PRODUCT_XY, axis=0),),
          expected=(coordax.SelectedAxis(PRODUCT_XY, axis=0),),
      ),
      dict(
          testcase_name='selections_with_following',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
              coordax.NamedAxis('z', 4),
          ),
          expected=(
              PRODUCT_XY,
              coordax.NamedAxis('z', 4),
          ),
      ),
      dict(
          testcase_name='selections_with_preceeding',
          coordinates=(
              coordax.NamedAxis('z', 4),
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
          ),
          expected=(
              coordax.NamedAxis('z', 4),
              PRODUCT_XY,
          ),
      ),
      dict(
          testcase_name='selections_split',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.NamedAxis('z', 4),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
          ),
          expected=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.NamedAxis('z', 4),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
          ),
      ),
      dict(
          testcase_name='two_selected_axes_consolidate_after',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.SelectedAxis(coordax.NamedAxis('x', 4), axis=0),
          ),
          expected=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.NamedAxis('x', 4),
          ),
      ),
      dict(
          testcase_name='two_selected_axes_consolidate_before',
          coordinates=(
              coordax.SelectedAxis(coordax.NamedAxis('x', 4), axis=0),
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
          ),
          expected=(
              coordax.NamedAxis('x', 4),
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
          ),
      ),
  )
  def test_consolidate_coordinates(self, coordinates, expected):
    actual = core.consolidate_coordinates(*coordinates)
    self.assertEqual(actual, expected)

  @parameterized.named_parameters(
      dict(
          testcase_name='selected_axes_compoents_merge',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.SelectedAxis(PRODUCT_XY, axis=1),
          ),
          expected=PRODUCT_XY,
      ),
      dict(
          testcase_name='selected_axis_simplified',
          coordinates=(
              coordax.SelectedAxis(coordax.NamedAxis('x', 4), axis=0),
              coordax.NamedAxis('z', 7),
          ),
          expected=coordax.CartesianProduct(
              (coordax.NamedAxis('x', 4), coordax.NamedAxis('z', 7))
          ),
      ),
      dict(
          testcase_name='cartesian_product_unraveled',
          coordinates=(
              coordax.NamedAxis('x', 7),
              coordax.CartesianProduct(
                  (coordax.NamedAxis('y', 7), coordax.NamedAxis('z', 4))
              ),
          ),
          expected=coordax.CartesianProduct((
              coordax.NamedAxis('x', 7),
              coordax.NamedAxis('y', 7),
              coordax.NamedAxis('z', 4),
          )),
      ),
      dict(
          testcase_name='consolidate_over_parts',
          coordinates=(
              coordax.SelectedAxis(PRODUCT_XY, axis=0),
              coordax.CartesianProduct((
                  coordax.SelectedAxis(PRODUCT_XY, axis=1),
                  coordax.NamedAxis('z', 4)
              )),
          ),
          expected=coordax.CartesianProduct((
              coordax.NamedAxis('x', 2),
              coordax.NamedAxis('y', 3),
              coordax.NamedAxis('z', 4),
          )),
      ),
  )
  def test_compose_coordinates(self, coordinates, expected):
    actual = core.compose_coordinates(*coordinates)
    self.assertEqual(actual, expected)

  @parameterized.named_parameters(
      dict(
          testcase_name='view_with_name',
          array=np.arange(5 * 3).reshape((5, 3)),
          tags=('i', 'j'),
          untags=('i',),
          expected_dims=(0, 'j'),
          expected_named_shape={'j': 3},
          expected_positional_shape=(5,),
          expected_coord_field_keys=set(),
      ),
      dict(
          testcase_name='view_with_name_and_coord',
          array=np.arange(5 * 3).reshape((5, 3, 1)),
          tags=('i', 'j', coordax.LabeledAxis('k', np.arange(1))),
          untags=('j',),
          expected_dims=('i', 0, 'k'),
          expected_named_shape={'i': 5, 'k': 1},
          expected_positional_shape=(3,),
          expected_coord_field_keys=set(['k']),
      ),
      dict(
          testcase_name='tag_prefix',
          array=np.arange(5 * 3).reshape((5, 3, 1)),
          tags=(coordax.LabeledAxis('k', np.arange(5)), 'j'),
          untags=(),  # untag can only be used on fully labeled fields.
          expected_dims=('k', 'j', 0),
          expected_named_shape={'k': 5, 'j': 3},
          expected_positional_shape=(1,),
          expected_coord_field_keys=set(['k']),
      ),
  )
  def test_field_properties(
      self,
      array: np.ndarray,
      tags: tuple[str | coordax.Coordinate, ...],
      untags: tuple[str | coordax.Coordinate, ...],
      expected_dims: tuple[str | int, ...],
      expected_named_shape: dict[str, int],
      expected_positional_shape: tuple[int, ...],
      expected_coord_field_keys: set[str],
  ):
    """Tests that field properties are correctly set."""
    field = coordax.wrap(array)
    if len(tags) == array.ndim:
      field = field.tag(*tags)
    else:
      field = field.tag_prefix(*tags)
    if untags:
      field = field.untag(*untags)
    testing.assert_field_properties(
        actual=field,
        data=array,
        dims=expected_dims,
        named_shape=expected_named_shape,
        positional_shape=expected_positional_shape,
        coord_field_keys=expected_coord_field_keys,
    )
    with self.subTest('broadcast_to_positional'):
      new_positional_shape = (7,) + field.positional_shape
      broadcasted = field.broadcast_to(positional_shape=new_positional_shape)
      testing.assert_field_properties(
          actual=broadcasted,
          named_shape=expected_named_shape,
          positional_shape=new_positional_shape,
          coord_field_keys=expected_coord_field_keys,
      )
      field.check_valid()

  @parameterized.named_parameters(
      dict(
          testcase_name='sum_simple',
          field_a=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          field_b=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          op=operator.add,
          expected_result=coordax.wrap(
              np.arange(2 * 3 * 4).reshape((2, 4, 3)) * 2
          ),
      ),
      dict(
          testcase_name='sum_aligned',
          field_a=coordax.wrap(np.arange(2 * 3).reshape((2, 3)), 'x', 'y'),
          field_b=coordax.wrap(
              np.arange(2 * 3)[::-1].reshape((3, 2)), 'y', 'x'
          ),
          op=operator.add,
          expected_result=coordax.wrap(
              np.array([[5, 4, 3], [7, 6, 5]]), 'x', 'y'
          ),
      ),
      dict(
          testcase_name='product_aligned',
          field_a=coordax.wrap(np.arange(2 * 3).reshape((2, 3))).tag_prefix(
              'x'
          ),
          field_b=coordax.wrap(np.arange(2), 'x'),
          op=operator.mul,
          expected_result=coordax.wrap(
              np.arange(2 * 3).reshape((2, 3)) * np.array([[0], [1]])
          ).tag_prefix('x'),
      ),
  )
  def test_field_binary_ops(
      self,
      field_a: coordax.Field,
      field_b: coordax.Field,
      op: Callable[[coordax.Field, coordax.Field], coordax.Field],
      expected_result: coordax.Field,
  ):
    """Tests that field binary ops work as expected."""
    actual = op(field_a, field_b)
    actual.check_valid()
    testing.assert_fields_allclose(actual=actual, desired=expected_result)

  def test_field_repr(self):
    expected = (
        "Field(named_array=<NamedArray int64(| x:2, y:3) "
        "(wrapping numpy.ndarray)>, coords={'x': NamedAxis(name='x', size=2), "
        "'y': LabeledAxis(name='y', ticks=np.array([7, 8, 9]))})"
    )
    actual = coordax.wrap(
        np.array([[1, 2, 3], [4, 5, 6]]),
        'x',
        coordax.LabeledAxis('y', np.array([7, 8, 9])),
    )
    self.assertEqual(repr(actual), expected)

  @parameterized.named_parameters(
      dict(
          testcase_name='_name_&_name',
          array=np.arange(4),
          tags=('idx',),
          untags=('idx',),
      ),
      dict(
          testcase_name='coord_&_name',
          array=np.arange(4),
          tags=(coordax.NamedAxis('idx', 4),),
          untags=('idx',),
      ),
      dict(
          testcase_name='coord_&_coord',
          array=np.arange(4),
          tags=(coordax.NamedAxis('idx', 4),),
          untags=(coordax.NamedAxis('idx', 4),),
      ),
      dict(
          testcase_name='names_&_partial_name',
          array=np.arange(2 * 3).reshape((2, 3)),
          tags=('x', 'y'),
          untags=('x',),
          full_unwrap=False,
      ),
      dict(
          testcase_name='names_&_partial_coord',
          array=np.arange(2 * 3).reshape((2, 3)),
          tags=('x', 'y'),
          untags=(coordax.NamedAxis('y', 3),),
          full_unwrap=False,
      ),
      dict(
          testcase_name='names_&_coords',
          array=np.arange(2 * 3).reshape((2, 3)),
          tags=('x', 'y'),
          untags=(
              coordax.NamedAxis('x', 2),
              coordax.NamedAxis('y', 3),
          ),
          full_unwrap=True,
      ),
      dict(
          testcase_name='names_&_product_coord',
          array=np.arange(2 * 3).reshape((2, 3)),
          tags=('x', 'y'),
          untags=(
              coordax.compose_coordinates(
                  coordax.NamedAxis('x', 2),
                  coordax.NamedAxis('y', 3),
              ),
          ),
          full_unwrap=True,
      ),
      dict(
          testcase_name='product_coord_&_product_coord',
          array=np.arange(2 * 3).reshape((2, 3)),
          tags=(
              coordax.compose_coordinates(
                  coordax.NamedAxis('x', 2),
                  coordax.NamedAxis('y', 3),
              ),
          ),
          untags=(
              coordax.compose_coordinates(
                  coordax.NamedAxis('x', 2),
                  coordax.NamedAxis('y', 3),
              ),
          ),
          full_unwrap=True,
      ),
      dict(
          testcase_name='mixed_&_names',
          array=np.arange(2 * 3 * 4).reshape((2, 4, 3)),
          tags=('x', coordax.NamedAxis('y', 4), 'z'),
          untags=('y', 'z'),
          full_unwrap=False,
      ),
      dict(
          testcase_name='mixed_&_wrong_names',
          array=np.arange(2 * 3 * 4).reshape((2, 4, 3)),
          tags=('x', coordax.NamedAxis('y_prime', 4), 'z'),
          untags=('y', 'z'),
          full_unwrap=False,
          should_raise_on_untag=True,
      ),
      dict(
          testcase_name='coord_&_wrong_coord_value',
          array=np.arange(9),
          tags=(
              coordax.LabeledAxis(
                  'z',
                  np.arange(9),
              ),
          ),
          untags=(coordax.LabeledAxis('z', np.arange(9) + 1),),
          full_unwrap=False,
          should_raise_on_untag=True,
      ),
  )
  def test_tag_then_untag_by(
      self,
      array: np.ndarray,
      tags: tuple[str | coordax.Coordinate, ...],
      untags: tuple[str | coordax.Coordinate, ...],
      should_raise_on_untag: bool = False,
      full_unwrap: bool = True,
  ):
    """Tests that tag and untag on Field work as expected."""
    with self.subTest('tag'):
      field = coordax.wrap(array, *tags)
      expected_dims = sum(
          [
              tag.dims if isinstance(tag, coordax.Coordinate) else (tag,)
              for tag in tags
          ],
          start=tuple(),
      )
      chex.assert_trees_all_equal(field.dims, expected_dims)

    with self.subTest('untag'):
      if should_raise_on_untag:
        with self.assertRaises(ValueError):
          field.untag(*untags)
      else:
        untagged = field.untag(*untags)
        if full_unwrap:
          unwrapped = untagged.unwrap()
          np.testing.assert_array_equal(unwrapped, array)

  @parameterized.named_parameters(
      dict(
          testcase_name='full_by_names',
          f=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          tags=('x', 'y', 'z'),
          expected_dims=('x', 'y', 'z'),
          expected_coord_field_keys=set(),
      ),
      dict(
          testcase_name='partial_by_names',
          f=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          tags=('x', 'y'),
          expected_dims=('x', 'y', 0),
          expected_coord_field_keys=set(),
      ),
      dict(
          testcase_name='partial_by_name&coord',
          f=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          tags=('x', coordax.LabeledAxis('y', np.arange(4))),
          expected_dims=('x', 'y', 0),
          expected_coord_field_keys=set(['y']),
      ),
      dict(
          testcase_name='partial_by_product_coord',
          f=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          tags=(
              coordax.compose_coordinates(
                  coordax.LabeledAxis('x', np.arange(2)),
                  coordax.LabeledAxis('z', np.linspace(0, 1, 4)),
              ),
          ),
          expected_dims=('x', 'z', 0),
          expected_coord_field_keys=set(['x', 'z']),
      ),
  )
  def test_tag_prefix(
      self,
      f: coordax.Field,
      tags: tuple[str | tuple[str, ...] | coordax.Coordinate, ...],
      expected_dims: tuple[str | int, ...],
      expected_coord_field_keys: set[str],
  ):
    """Tests that tag_prefix works as expected."""
    tagged_f = f.tag_prefix(*tags)
    testing.assert_field_properties(
        actual=tagged_f,
        dims=expected_dims,
        coord_field_keys=expected_coord_field_keys,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='partial_by_name_tail',
          f=coordax.wrap(np.arange(2 * 3).reshape((2, 3)), 'x', 'y'),
          untags=('y',),
          prefix_expected_dims=('x', 0),
          suffix_expected_dims=('x', 0),
          expected_coord_field_keys=set(),
      ),
      dict(
          testcase_name='partial_by_name_head',
          f=coordax.wrap(np.arange(2 * 3).reshape((2, 3)), 'x', 'y'),
          untags=('x',),
          prefix_expected_dims=(0, 'y'),
          suffix_expected_dims=(0, 'y'),
          expected_coord_field_keys=set(),
      ),
      dict(
          testcase_name='partial_by_coord',
          f=coordax.wrap(
              np.arange(2 * 3 * 5).reshape((2, 3, 5)),
              'x',
              coordax.LabeledAxis('y', np.arange(3)),
              'z',
          ),
          untags=('x', 'z'),
          prefix_expected_dims=(0, 'y', 1),
          suffix_expected_dims=(0, 'y', 1),
          expected_coord_field_keys=set(['y']),
      ),
      dict(
          testcase_name='with_positional_prefix_partial_by_name',
          f=coordax.wrap(
              np.arange(2 * 3 * 5).reshape((2, 3, 5)),
              'x',
              coordax.LabeledAxis('y', np.arange(3)),
              'z',
          ).untag('x'),
          untags=('z',),
          prefix_expected_dims=(1, 'y', 0),
          suffix_expected_dims=(0, 'y', 1),
          expected_coord_field_keys=set(['y']),
      ),
      dict(
          testcase_name='mixed_partial_by_name',
          f=coordax.wrap(np.ones((1, 1, 1, 1)), 'x', 'y', 'z', 'w').untag(
              'y', 'w'
          ),
          untags=('z',),
          prefix_expected_dims=('x', 1, 0, 2),
          suffix_expected_dims=('x', 0, 2, 1),
          expected_coord_field_keys=set(),
      ),
  )
  def test_untag(
      self,
      f: coordax.Field,
      untags: tuple[str | tuple[str, ...] | coordax.Coordinate, ...],
      prefix_expected_dims: tuple[str | int, ...],
      suffix_expected_dims: tuple[str | int, ...],
      expected_coord_field_keys: set[str],
  ):
    """Tests that untag_prefix works as expected."""
    with self.subTest('prefix'):
      untagged_f = f.untag_prefix(*untags)
      testing.assert_field_properties(
          actual=untagged_f,
          dims=prefix_expected_dims,
          coord_field_keys=expected_coord_field_keys,
      )
    with self.subTest('suffix'):
      untagged_f = f.untag_suffix(*untags)
      testing.assert_field_properties(
          actual=untagged_f,
          dims=suffix_expected_dims,
          coord_field_keys=expected_coord_field_keys,
      )

  @parameterized.named_parameters(
      dict(
          testcase_name='cos(x)',
          inputs=coordax.wrap(np.arange(2 * 3 * 4).reshape((2, 4, 3))),
          tags=('x', 'y', 'z'),
          untags=('x',),
          fn=jnp.cos,
          expected_dims=('y', 'z', 0),
          expected_coord_field_keys=set(),
          result_fn=lambda x: jnp.cos(x).transpose((1, 2, 0)),
      ),
      dict(
          testcase_name='norm(x, z)',
          inputs=coordax.wrap(np.arange(2 * 3 * 5).reshape((2, 3, 5))),
          tags=('x', coordax.LabeledAxis('y', np.arange(3)), 'z'),
          fn=jnp.linalg.norm,
          untags=('x', 'z'),
          expected_dims=('y',),
          expected_coord_field_keys=set(['y']),
          result_fn=lambda x: jnp.linalg.norm(x, axis=(0, 2)),
      ),
  )
  def test_cmap(
      self,
      inputs: coordax.Field,
      tags: tuple[str | coordax.Coordinate, ...],
      untags: tuple[str | coordax.Coordinate, ...],
      fn: Callable[..., Any],
      expected_dims: tuple[str | int, ...],
      expected_coord_field_keys: set[str],
      result_fn: Callable[..., Any],
  ):
    """Tests that cmap works as expected."""
    input_f = inputs.tag_prefix(*tags).untag(*untags)
    actual = coordax.cmap(fn)(input_f)
    expected_values = result_fn(inputs.data)
    testing.assert_field_properties(
        actual=actual,
        data=expected_values,
        dims=expected_dims,
        shape=expected_values.shape,
        coord_field_keys=expected_coord_field_keys,
    )

  def test_jax_transforms(self):
    """Tests that vmap/scan work with Field with leading positional axes."""
    coords = coordax.LabeledAxis('x', np.array([2, 3, 7]))
    batch, length = 4, 10
    vmap_axis = coordax.NamedAxis('i', batch)
    scan_axis = coordax.LabeledAxis('timedelta', np.arange(length))

    def initialize(data):
      return coordax.wrap(data, coords)

    def body_fn(c, _):
      return (c + 1, c)

    with self.subTest('scan'):
      data = np.zeros(coords.shape)
      init = initialize(data)
      _, scanned = jax.lax.scan(body_fn, init, length=length)
      scanned = scanned.tag_prefix(scan_axis)
      testing.assert_field_properties(
          actual=scanned,
          dims=('timedelta', 'x'),
          shape=(length,) + coords.shape,
      )

    with self.subTest('vmap'):
      batch_data = np.zeros(vmap_axis.shape + coords.shape)
      batch_init = jax.vmap(initialize)(batch_data).tag(vmap_axis)
      testing.assert_field_properties(
          batch_init, dims=('i', 'x'), shape=batch_data.shape
      )

    with self.subTest('vmap_of_scan'):
      batch_data = np.zeros(vmap_axis.shape + coords.shape)
      batch_init = jax.vmap(initialize)(batch_data).tag(vmap_axis)
      # we ensure that inputs to scan/vmap have positional prefix, since
      # otherwise shape changes are not allowed.
      batch_init = batch_init.untag(vmap_axis).with_positional_prefix()
      scan_fn = functools.partial(jax.lax.scan, body_fn, length=length)
      _, scanned = jax.vmap(scan_fn, in_axes=0)(batch_init)
      scanned = scanned.tag_prefix(vmap_axis, scan_axis)
      testing.assert_field_properties(
          actual=scanned,
          dims=('i', 'timedelta', 'x'),
          shape=(batch, length) + coords.shape,
      )

    with self.subTest('scan_of_vmap'):
      batch_data = np.zeros(vmap_axis.shape + coords.shape)
      batch_init = jax.vmap(initialize)(batch_data).tag(vmap_axis)
      # we ensure that inputs to scan/vmap have positional prefix, since
      # otherwise shape changes are not allowed.
      batch_init = batch_init.untag(vmap_axis).with_positional_prefix()
      vmaped_body_fn = jax.vmap(body_fn)
      scan_fn = functools.partial(jax.lax.scan, vmaped_body_fn, length=length)
      _, scanned = scan_fn(batch_init)
      scanned = scanned.tag_prefix(scan_axis, vmap_axis)
      testing.assert_field_properties(
          actual=scanned,
          dims=('timedelta', 'i', 'x'),
          shape=(length, batch) + coords.shape,
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for interpolation routines."""

from absl.testing import absltest
from absl.testing import parameterized
import jax
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import interpolators
import numpy as np


class InterpolatorsTest(parameterized.TestCase):
  """Tests interpolator modules."""

  @parameterized.named_parameters(
      dict(
          testcase_name='spherical_coords',
          target_grid=coordinates.SphericalHarmonicGrid.T21(),
          input_coords=coordinates.SphericalHarmonicGrid.TL31(),
      ),
  )
  def test_spectral_regridder(self, target_grid, input_coords):
    regridder = interpolators.SpectralRegridder(target_grid)
    inputs = cx.wrap(np.ones(input_coords.shape), input_coords)
    outputs = regridder(inputs)
    output_coords = cx.compose_coordinates(*outputs.coords.values())
    self.assertEqual(output_coords, target_grid)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()
